/*
 * All content copyright (c) 2003-2008 Terracotta, Inc., except as may otherwise be noted in a separate copyright
 * notice. All rights reserved.
 */
package com.tc.object;

import com.tc.aspectwerkz.reflect.ClassInfo;
import com.tc.aspectwerkz.reflect.impl.java.JavaClassInfo;
import com.tc.exception.TCRuntimeException;
import com.tc.object.applicator.AccessibleObjectApplicator;
import com.tc.object.applicator.ArrayApplicator;
import com.tc.object.applicator.CalendarApplicator;
import com.tc.object.applicator.ChangeApplicator;
import com.tc.object.applicator.FileApplicator;
import com.tc.object.applicator.LiteralTypesApplicator;
import com.tc.object.applicator.PhysicalApplicator;
import com.tc.object.applicator.ProxyApplicator;
import com.tc.object.config.DSOClientConfigHelper;
import com.tc.object.config.TransparencyClassSpec;
import com.tc.object.dna.api.DNAEncoding;
import com.tc.object.field.TCFieldFactory;
import com.tc.object.loaders.ClassProvider;
import com.tc.object.loaders.LoaderDescription;
import com.tc.util.ClassUtils;
import com.tc.util.runtime.Vm;

import java.io.File;
import java.lang.reflect.Constructor;
import java.util.Calendar;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentMap;

/**
 * @author steve
 */
public class TCClassFactoryImpl implements TCClassFactory {
  private static final boolean                   IS_IBM                    = Vm.isIBM();

  private static Class[]                         APPLICATOR_CSTR_SIGNATURE = new Class[] { DNAEncoding.class };

  private final ConcurrentMap<Class<?>, TCClass> classes                   = new ConcurrentHashMap<Class<?>, TCClass>();
  private final TCFieldFactory                   fieldFactory;
  private final DSOClientConfigHelper            config;
  private final ClassProvider                    classProvider;
  private final DNAEncoding                      encoding;

  public TCClassFactoryImpl(final TCFieldFactory fieldFactory, final DSOClientConfigHelper config,
                            final ClassProvider classProvider, final DNAEncoding dnaEncoding) {
    this.fieldFactory = fieldFactory;
    this.config = config;
    this.classProvider = classProvider;
    this.encoding = dnaEncoding;
  }

  public TCClass getOrCreate(final Class clazz, final ClientObjectManager objectManager) {
    TCClass rv = classes.get(clazz);
    if (rv != null) return rv;

    LoaderDescription loaderDesc = classProvider.getLoaderDescriptionFor(clazz);
    String className = clazz.getName();
    ClassInfo classInfo = JavaClassInfo.getClassInfo(clazz);
    rv = new TCClassImpl(fieldFactory, this, objectManager, config.getTCPeerClass(clazz),
                         getLogicalSuperClassWithDefaultConstructor(clazz), loaderDesc, config
                             .getLogicalExtendingClassName(className), config.isLogical(className), config
                             .isCallConstructorOnLoad(classInfo), config.hasOnLoadInjection(classInfo), config
                             .getOnLoadScriptIfDefined(classInfo), config.getOnLoadMethodIfDefined(classInfo), config
                             .isUseNonDefaultConstructor(clazz), config.useResolveLockWhenClearing(clazz), config
                             .getPostCreateMethodIfDefined(className), config.getPreCreateMethodIfDefined(className));

    TCClass existing = classes.putIfAbsent(clazz, rv);
    return existing == null ? rv : existing;
  }

  public Class getLogicalSuperClassWithDefaultConstructor(Class clazz) {
    TransparencyClassSpec spec = config.getSpec(clazz.getName());
    if (spec == null) { return null; }

    while (clazz != null) {
      if (spec == null) { return null; }
      if (spec.isLogical()) {
        Constructor c = null;
        try {
          c = clazz.getDeclaredConstructor(new Class[0]);
        } catch (SecurityException e) {
          throw new TCRuntimeException(e);
        } catch (NoSuchMethodException e) {
          // c is already null
        }
        if (c != null) { return clazz; }
      }
      clazz = clazz.getSuperclass();
      if (clazz != null) {
        spec = config.getSpec(clazz.getName());
      } else {
        spec = null;
      }
    }
    return null;
  }

  public ChangeApplicator createApplicatorFor(final TCClass clazz, final boolean indexed) {
    if (indexed) { return new ArrayApplicator(encoding); }
    String name = clazz.getName();
    Class applicatorClazz = config.getChangeApplicator(clazz.getPeerClass());

    if (applicatorClazz == null) {
      if (LiteralValues.isLiteral(name)) {
        return new LiteralTypesApplicator(clazz, encoding);
      } else if (clazz.isProxyClass()) {
        return new ProxyApplicator(encoding);
      } else if (ClassUtils.isPortableReflectionClass(clazz.getPeerClass())) {
        return new AccessibleObjectApplicator(encoding);
      } else if (File.class.isAssignableFrom(clazz.getPeerClass())) {
        return new FileApplicator(clazz, encoding);
      } else if (Calendar.class.isAssignableFrom(clazz.getPeerClass())) {
        return new CalendarApplicator(clazz, encoding);
      } else if (IS_IBM && "java.util.concurrent.atomic.AtomicInteger".equals(name)) {
        try {
          Class klass = Class.forName("com.tc.object.applicator.AtomicIntegerApplicator");
          Constructor constructor = klass.getDeclaredConstructor(APPLICATOR_CSTR_SIGNATURE);
          return (ChangeApplicator) constructor.newInstance(new Object[] { encoding });
        } catch (Exception e) {
          throw new AssertionError(e);
        }
      } else if (IS_IBM && "java.util.concurrent.atomic.AtomicLong".equals(name)) {
        try {
          Class klass = Class.forName("com.tc.object.applicator.AtomicLongApplicator");
          Constructor constructor = klass.getDeclaredConstructor(APPLICATOR_CSTR_SIGNATURE);
          return (ChangeApplicator) constructor.newInstance(new Object[] { encoding });
        } catch (Exception e) {
          throw new AssertionError(e);
        }
      } else {
        return new PhysicalApplicator(clazz, encoding);
      }
    }

    try {
      Constructor cstr = applicatorClazz.getConstructor(APPLICATOR_CSTR_SIGNATURE);
      Object[] params = new Object[] { encoding };
      return (ChangeApplicator) cstr.newInstance(params);
    } catch (Exception e) {
      throw new AssertionError(e);
    }
  }
}
